# coding: utf-8
import logging
import time
from typing import Union, Dict, List, Tuple

import numpy as np
from PIL import Image

from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import WebDriverException, TimeoutException, NoSuchElementException
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.common.by import By
from appium.webdriver import WebElement
from appium.webdriver.common.mobileby import MobileBy
from appium.webdriver.common.appiumby import AppiumBy

from appium_extended.appium_base import AppiumBase


class AppiumGet(AppiumBase):
    """
    Класс расширяющий Appium.
    Обеспечивает получение чего-либо со страницы.
    """

    def __init__(self, logger: logging.Logger):
        super().__init__(logger=logger)

    def _get_element(self,
                     locator: Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str] = None,
                     by: Union[MobileBy, AppiumBy, By, str] = None,
                     value: Union[str, Dict, None] = None,
                     timeout_elem: int = 10,
                     timeout_method: int = 600,
                     elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                     contains: bool = False
                     ) -> \
            Union[WebElement, None]:
        """
        Метод обеспечивает поиск элемента в текущей DOM структуре.
        Должен принимать либо локатор, либо значения by и value.

        Usage:
            element = app._get_element(locator=("id", "foo")).
            element = app._get_element(element).
            element = app._get_element(locator={'text': 'foo'}).
            element = app._get_element(locator='/path/to/file/pay_agent.png').
            element = app._get_element(locator=part_image, elements_range={'class':'android.widget.FrameLayout', 'package':'ru.app.debug'}).
            element = app._get_element(by="id", value="ru.sigma.app.debug:id/backButton").
            element = app._get_element(by=MobileBy.ID, value="ru.sigma.app.debug:id/backButton").
            element = app._get_element(by=AppiumBy.ID, value="ru.sigma.app.debug:id/backButton").
            element = app._get_element(by=By.ID, value="ru.sigma.app.debug:id/backButton").

        Args:
            locator: tuple / WebElement / dict / str, определяет локатор элемента.
                tuple - локатор в виде ('атрибут', 'значение')
                WebElement - объект веб элемента
                dict - словарь, содержащий пары атрибут: значение
                str - путь до файла с изображением элемента.
            by: MobileBy, AppiumBy, By, str, тип локатора для поиска элемента (всегда в связке с value)
            value: str, dict, None, значение локатора или словарь аргументов, если используется AppiumBy.XPATH.
            timeout_elem: int, время ожидания элемента.
            timeout_method: int, время ожидания метода поиска элемента.
            elements_range: tuple, list, dict, None, ограничивает поиск элемента в указанном диапазоне
            (для поиска по изображению).
            contains: для поиска по dict, ищет элемент содержащий фрагмент значения

        Returns:
            WebElement или None, если элемент не был найден.
        """
        # Проверка и подготовка аргументов
        if (not locator) and (not by or not value):
            self.logger.error(f"Некорректные аргументы!\n"
                              f"{locator=}\n"
                              f"{by=}\n"
                              f"{value=}\n"
                              f"{timeout_elem=}\n")
            return None
        if not locator and (by and value):
            locator = (by, value)
        if locator is None:
            return None

        # Объявление стратегии поиска элементов
        locator_handler = {
            # возвращает себя же
            WebElement: self.helper.handle_webelement_locator,
            # возвращает себя же
            'WebElementExtended': self.helper.handle_webelement_locator,
            # составляет локатор типа tuple из словаря с атрибутами искомого элемента
            dict: self.helper.handle_dict_locator,
            # производит поиск элементов по фрагменту изображения, возвращает список элементов
            str: self.helper.handle_string_locator,
        }

        # Цикл подготовки локатора и поиска элементов
        start_time = time.time()
        while not isinstance(locator, WebElement) and time.time() - start_time < timeout_method:
            # Выявление типа данных локатора для его подготовки
            locator_type = type(locator)
            # Если локатор типа tuple, то выполняется извлечение элементов
            if isinstance(locator, tuple):
                wait = WebDriverWait(driver=self.driver, timeout=timeout_elem)
                try:
                    element = wait.until(EC.presence_of_element_located(locator))
                    return element
                except NoSuchElementException:
                    return None
                except TimeoutException:
                    # self.logger.error(f"Элемент не обнаружен!\n"
                    #                   f"{locator=}\n"
                    #                   f"{timeout_elem=}\n\n" +
                    #                   "{}\n".format(e))
                    # self.logger.error("page source ", self.driver.page_source)
                    return None
                except WebDriverException:
                    # self.logger.error(f"Элемент не обнаружен!\n"
                    #                   f"{locator=}\n"
                    #                   f"{timeout_elem=}\n\n" +
                    #                   "{}\n".format(e))
                    # self.logger.error("page source ", self.driver.page_source)
                    return None
            # Выполнение подготовки локатора
            handler = locator_handler.get(locator_type)
            if locator is None:
                return None
            locator = handler(locator=locator, timeout=timeout_elem, elements_range=elements_range, contains=contains)
        # Подбирает результат после поиска по изображению
        if isinstance(locator, WebElement):
            return locator
        self.logger.error(f"Что-то пошло не так\n"
                          f"{locator=}\n"
                          f"{by=}\n"
                          f"{value=}\n"
                          f"{timeout_elem=}\n"
                          f"{timeout_method=}\n")
        return None

    def _get_elements(self,
                      locator: Union[Tuple, List[WebElement], Dict[str, str], str] = None,
                      by: Union[MobileBy, AppiumBy, By, str] = None,
                      value: Union[str, Dict, None] = None,
                      timeout_elements: int = 10,
                      timeout_method: int = 600,
                      elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                      contains: bool = True) -> \
            Union[List[WebElement], None]:
        """
        Метод обеспечивает поиск элементов в текущей DOM структуре.
        Должен принять либо локатор, либо by и value.
        При locator:str настоятельно рекомендуется использовать диапазон поиска elements_range.

        Usage:
            elements = app.get_elements(locator=("id", "foo")).
            elements = app.get_elements(locator={'text': 'foo'}).
            elements = app.get_elements(locator='/path/to/file/pay_agent.png').
            elements = app.get_elements(by="id", value="ru.sigma.app.debug:id/backButton").
            elements = app.get_elements(by=MobileBy.ID, value="ru.sigma.app.debug:id/backButton").
            elements = app.get_elements(by=AppiumBy.ID, value="ru.sigma.app.debug:id/backButton").
            elements = app.get_elements(by=By.ID, value="ru.sigma.app.debug:id/backButton").

        Args:
            locator: tuple or WebElement or Dict[str, str], str, локатор tuple или Веб Элемент или словарь {'атрибут': 'значение'} или str как путь до файла с изображением элемента.
            by:[MobileBy, AppiumBy, By, str], тип локатора для поиска элемента (всегда в связке с value)
            value: Union[str, Dict, None], значение локатора или словарь аргументов, если используется AppiumBy.XPATH
            timeout_elements: #TODO fill me
            timeout_method: #TODO fill me
            elements_range: #TODO fill me

        Returns:
            Список WebElement'ов, или пустой список в случае их отсутствия.
        """
        # Проверка и подготовка аргументов
        if not locator and (not by or not value):
            self.logger.error(f"Некорректные аргументы!\n"
                              f"{locator=}\n"
                              f"{by=}\n"
                              f"{value=}\n"
                              f"{timeout_elements=}\n"
                              f"{timeout_method=}\n")
            return None
        if not locator and (by and value):
            locator = (by, value)
        if locator is None:
            return None

        # Объявление стратегии поиска элементов
        locator_handler = {
            # подразумевается список элементов, возвращает себя же
            list: self.helper.handle_webelement_locator_elements,
            # составляет локатор типа tuple из словаря с атрибутами искомого элемента
            dict: self.helper.handle_dict_locator_elements,
            # производит поиск элементов по фрагменту изображения, возвращает список элементов
            str: self.helper.handle_string_locator_elements,
        }

        # Цикл подготовки локатора и поиска элементов
        start_time = time.time()
        while not isinstance(locator, list) and time.time() - start_time < timeout_method:
            # Выявление типа данных локатора для его подготовки
            locator_type = type(locator)
            # Если локатор типа tuple, то выполняется извлечение элементов
            if isinstance(locator, tuple):
                wait = WebDriverWait(driver=self.driver, timeout=timeout_elements)
                try:
                    element = wait.until(EC.presence_of_all_elements_located(locator))
                    return element
                except WebDriverException:
                    # self.logger.error(f"Элемент не обнаружен!\n"
                    #                   f"{locator=}\n"
                    #                   f"{by=}\n"
                    #                   f"{value=}\n"
                    #                   f"{timeout_elements=}\n"
                    #                   f"{timeout_method=}\n\n" +
                    #                   "{}\n".format(e))
                    return None
            # Выполнение подготовки локатора
            handler = locator_handler.get(locator_type)
            locator = handler(locator=locator,
                              timeout=timeout_elements,
                              elements_range=elements_range,
                              contains=contains)
        # Подбирает результат после поиска по изображению
        if isinstance(locator, list):
            return locator
        self.logger.error(f"\nЧто-то пошло не так\n"
                          f"{locator=}\n"
                          f"{by=}\n"
                          f"{value=}\n"
                          f"{timeout_elements=}\n"
                          f"{timeout_method=}\n")
        return None

    def _get_image_coordinates(self,
                               image: Union[bytes, np.ndarray, Image.Image, str],
                               full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                               threshold: float = 0.7,
                               ) -> Union[Tuple[int, int, int, int], None]:
        return self.helper.get_image_coordinates(image=image, full_image=full_image, threshold=threshold)

    def _get_inner_image_coordinates(self,
                                     outer_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                     inner_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                     threshold: float = 0.9) -> \
            Union[Tuple[int, int, int, int], None]:
        return self.helper.get_inner_image_coordinates(outer_image_path=outer_image_path,
                                                      inner_image_path=inner_image_path,
                                                      threshold=threshold)

    def _get_text_coordinates(self,
                              text: str,
                              language: str = 'rus',
                              image: Union[bytes, str, Image.Image, np.ndarray] = None, ) -> Tuple[int, int, int, int]:
        return self.helper.get_text_coordinates(text=text, language=language, image=image)

    def _get_screenshot_as_base64_decoded(self):
        return self.helper.get_screenshot_as_base64_decoded()



# coding: utf-8
import base64
import logging
import os
from typing import Union, List, Tuple, Optional

import cv2
import numpy as np
from PIL import Image
from pytesseract import pytesseract

from selenium.common.exceptions import WebDriverException

from appium_extended_helpers import helpers_decorators
from appium_extended_terminal.terminal import Terminal


class AppiumImage:
    """
    Класс работы с Appium.
    Обеспечивает работу с изображениями
    """

    def __init__(self, driver, logger: logging.Logger):
        self.logger = logger
        self.driver = driver
        self.terminal = Terminal(driver=self.driver, logger=logger)


    def get_screenshot_as_base64_decoded(self):
        screenshot = self.driver.get_screenshot_as_base64().encode('utf-8')
        screenshot = base64.b64decode(screenshot)
        return screenshot

    @helpers_decorators.retry
    def get_image_coordinates(self,
                              image: Union[bytes, np.ndarray, Image.Image, str],
                              full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                              threshold: Optional[float] = 0.7,
                              ) -> Union[Tuple[int, int, int, int], None]:
        """
        Находит координаты наиболее вероятного совпадения частичного изображения в полном изображении.
        Предназначен для поиска координат наиболее вероятного совпадения частичного изображения в полном изображении.

        Args:
            image (Union[bytes, np.ndarray, Image.Image, str]): путь к файлу частичного изображения, которое
            нужно найти внутри полного изображения.
            full_image (Union[bytes, np.ndarray, Image.Image, str]): путь к файлу полного изображения.
            threshold (float, optional): минимальный порог совпадения, необходимый для считывания совпадения допустимым.
                По умолчанию равно 0.7.

        Returns:
            tuple or None: кортеж с координатами наиболее вероятного совпадения (x1, y1, x2, y2) или None, если совпадение не найдено.
            :param threshold:
            :param full_image:
        """
        if full_image is None:
            screenshot = self.get_screenshot_as_base64_decoded()
            big_image = self.to_ndarray(image=screenshot, grayscale=True)
        else:
            big_image = self.to_ndarray(image=full_image, grayscale=True)  # Загрузка полного изображения

        small_image = self.to_ndarray(image=image, grayscale=True)  # Загрузка частичного изображения

        result = cv2.matchTemplate(big_image, small_image, cv2.TM_CCOEFF_NORMED)  # Поиск совпадений методом шаблона

        _, max_val, _, max_loc = cv2.minMaxLoc(
            result)  # Получение наименьшего и наибольшего значения, а также соответствующих координат

        if not max_val >= threshold:  # Если наибольшее значение совпадения не превышает порога, возвращаем None
            self.logger.error("find_coordinates_by_image(): Совпадений не найдено")
            return None

        # Вычисляем координаты левого верхнего и правого нижнего углов найденного совпадения
        left = int(max_loc[0])
        top = int(max_loc[1])
        width = small_image.shape[1]
        height = small_image.shape[0]
        right = left + width
        bottom = top + height

        return int(left), int(top), int(right), int(bottom)  # Возвращаем координаты наиболее вероятного совпадения

    @helpers_decorators.retry
    def get_inner_image_coordinates(self,
                                    outer_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                    inner_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                    threshold: float = 0.9) -> Union[Tuple[int, int, int, int], None]:
        """
        Находит изображение на экране и внутри него находит другое изображение (внутреннее).
        Предназначен для поиска изображения на экране и затем нахождения другого изображения (внутреннего)
        внутри этого обнаруженного изображения.

        Параметры:
            outer_image_path (str): Путь к файлу с изображением, которое нужно найти на экране.
            inner_image_path (str): Путь к файлу с изображением, которое нужно найти внутри внешнего изображения.
            threshold (float, optional): Пороговое значение сходства для шаблонного сопоставления. По умолчанию 0.9.

        Возвращает:
            tuple: Координаты внутреннего изображения относительно экрана в формате ((x1, y1), (x2, y2)).
                   Если внутреннее изображение не найдено, возвращает None.
        """
        # Получаем разрешение экрана
        screen_width, screen_height = self.terminal.get_screen_resolution()

        # Захватываем скриншот
        screenshot = base64.b64decode(self.driver.get_screenshot_as_base64())

        # Читаем скриншот
        full_image = self.to_ndarray(image=screenshot, grayscale=True)

        # Прочитать внешнее изображение
        outer_image = self.to_ndarray(image=outer_image_path, grayscale=True)

        # Прочитать внутреннее изображение
        inner_image = self.to_ndarray(image=inner_image_path, grayscale=True)

        # Вычисляем коэффициенты масштабирования
        width_ratio = screen_width / full_image.shape[1]
        height_ratio = screen_height / full_image.shape[0]

        # ...
        inner_image = cv2.resize(inner_image, None, fx=width_ratio, fy=height_ratio)
        outer_image = cv2.resize(outer_image, None, fx=width_ratio, fy=height_ratio)

        # Применить шаблонное сопоставление для поиска внешнего изображения
        outer_result = cv2.matchTemplate(full_image, outer_image, cv2.TM_CCOEFF_NORMED)

        # Найти максимальное значение сходства и его положение для внешнего изображения
        _, outer_max_val, _, outer_max_loc = cv2.minMaxLoc(outer_result)

        # Проверить, превышает ли максимальное значение сходства для внешнего изображения пороговое значение
        if outer_max_val >= threshold:
            # Получить размеры внешнего изображения
            outer_height, outer_width = outer_image.shape

            # Вычислить координаты внешнего изображения на экране
            outer_top_left = outer_max_loc
            outer_bottom_right = (outer_top_left[0] + outer_width, outer_top_left[1] + outer_height)

            # Извлечь область интереса (ROI), содержащую внешнее изображение
            outer_roi = full_image[outer_top_left[1]:outer_bottom_right[1], outer_top_left[0]:outer_bottom_right[0]]

            # Применить шаблонное сопоставление для поиска внутреннего изображения внутри ROI
            inner_result = cv2.matchTemplate(outer_roi, inner_image, cv2.TM_CCOEFF_NORMED)

            # Найти максимальное значение сходства и его положение для внутреннего изображения внутри ROI
            _, inner_max_val, _, inner_max_loc = cv2.minMaxLoc(inner_result)

            # Проверить, превышает ли максимальное значение сходства для внутреннего изображения пороговое значение
            if inner_max_val >= threshold:
                # Получить размеры внутреннего изображения
                inner_height, inner_width = inner_image.shape

                # Вычислить координаты внутреннего изображения относительно экрана
                inner_top_left = (outer_top_left[0] + inner_max_loc[0], outer_top_left[1] + inner_max_loc[1])
                inner_bottom_right = (inner_top_left[0] + inner_width, inner_top_left[1] + inner_height)

                # Вернуть координаты внутреннего изображения относительно экрана
                return inner_top_left + inner_bottom_right

        # Вернуть None, если внутреннее изображение не найдено
        return None

    def is_image_on_the_screen(self,
                               image: Union[bytes, np.ndarray, Image.Image, str],
                               threshold: float = 0.9) -> bool:
        """
        Сравнивает, присутствует ли заданное изображение на экране.

        Args:
            image: Строка, содержащая имя файла частичного изображения для поиска.
            threshold: Пороговое значение схожести части изображения со снимком экрана

        Returns:
            Логическое значение, указывающее, было ли частичное изображение найдено на экране.
        """
        try:
            screenshot = self.get_screenshot_as_base64_decoded()

            # Чтение снимка экрана и частичного изображения
            full_image = self.to_ndarray(image=screenshot, grayscale=True)
            small_image = self.to_ndarray(image=image, grayscale=True)

            # Проверка размеров изображений
            if small_image.shape[0] > full_image.shape[0] or small_image.shape[1] > full_image.shape[1]:
                self.logger.error("Частичное изображение больше снимка экрана.")
                return False

            # Сопоставление частичного изображения и снимка экрана
            result = cv2.matchTemplate(full_image, small_image, cv2.TM_CCOEFF_NORMED)

            # Извлечение коэффициента схожести и координат схожего участка
            _, max_val, _, _ = cv2.minMaxLoc(result)

            # Логирование
            self.logger.debug(f"Коэффициент схожести изображения: {max_val}")

            # Сравнение коэффициента схожести и порогового значения
            return max_val >= threshold

        except cv2.error as e:
            self.logger.error(f"is_image_on_the_screen(): {e}")
            return False
        except AssertionError as e:
            self.logger.error(f"is_image_on_the_screen(): {e}")
            return False
        except Exception as e:
            self.logger.error(f"is_image_on_the_screen(): {e}")
            return False

    def is_text_on_ocr_screen(self,
                              text: str,
                              screen: Union[bytes, np.ndarray, Image.Image, str] = None,
                              language: str = 'rus') -> bool:
        """
        Проверяет, присутствует ли заданный текст на экране.
        Распознавание текста производит с помощью библиотеки pytesseract.

        Аргументы:
        - text (str): Текст, который нужно найти на экране.
        - screen (bytes, optional): Скриншот в формате bytes. Если не указан, будет захвачен скриншот с помощью `self.driver`.
        - language (str): Язык распознавания текста. Значение по умолчанию: 'rus'.

        Возвращает:
        - bool: True, если заданный текст найден на экране. False в противном случае.
        """
        try:
            if screen is None:
                screenshot = self.get_screenshot_as_base64_decoded()
                image = self.to_ndarray(screenshot)
            else:
                image = self.to_ndarray(screen)

            # Бинаризация изображения
            _, image_bin = cv2.threshold(image, 0, 255,
                                         cv2.THRESH_BINARY | cv2.THRESH_OTSU)  # Применение бинаризации для получения двоичного изображения

            # Преобразование двоичного изображения в текст
            custom_config = r'--oem 3 --psm 6'
            ocr_text = pytesseract.image_to_string(image_bin, lang=language, config=custom_config)

            # Проверка наличия заданного текста в распознанном тексте
            return text.lower() in ocr_text.lower()

        except cv2.error as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False
        except pytesseract.TesseractError as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False
        except AssertionError as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False
        except Exception as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False

    @helpers_decorators.retry
    def get_many_coordinates_of_image(self,
                                      image: Union[bytes, np.ndarray, Image.Image, str],
                                      full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                                      cv_threshold: float = 0.7,
                                      coord_threshold: int = 5) -> Union[List[Tuple], None]:
        """
        Находит все вхождения частичного изображения внутри полного изображения.

        Args:
            full_image (str): путь к файлу полного изображения.
            image (str): путь к файлу частичного изображения, которое нужно найти внутри полного изображения.
            cv_threshold (float, optional): минимальный порог совпадения, необходимый для считывания совпадения допустимым.
                По умолчанию равно 0.7.
            coord_threshold (int, optional): целое число, представляющее максимальное различие между значениями x и y двух
                кортежей, чтобы они считались слишком близкими друг к другу. По умолчанию равно 5.

        Returns:
            list of tuples or None: список кортежей, содержащий расположение каждого найденного совпадения.
                Если совпадений не найдено, возвращается None.
        """

        if full_image is None:
            screenshot = self.get_screenshot_as_base64_decoded()
            big_image = self.to_ndarray(image=screenshot, grayscale=True)
        else:
            big_image = self.to_ndarray(image=full_image, grayscale=True)  # Загрузка полного изображения

        small_image = self.to_ndarray(image=image, grayscale=True)  # Загрузка частичного изображения

        # Найти частичное изображение в полном изображении
        result = cv2.matchTemplate(big_image, small_image, cv2.TM_CCOEFF_NORMED)  # Поиск совпадений методом шаблона

        # Получить все совпадения выше порога
        locations = np.where(result >= cv_threshold)  # Нахождение всех совпадений выше порога
        matches = list(zip(*locations[::-1]))  # Преобразование координат в список кортежей

        # Фильтрация слишком близких совпадений
        unique_list = []  # Создаем пустой список для хранения уникальных кортежей
        for (x1_coordinate, y1_coordinate) in matches:  # Итерируемся по списку кортежей
            exclude = False  # Инициализируем флаг exclude значением False
            for (x2_coordinate, y2_coordinate) in unique_list:  # Итерируемся по уникальным кортежам
                if abs(x1_coordinate - x2_coordinate) <= coord_threshold and abs(
                        y1_coordinate - y2_coordinate) <= coord_threshold:
                    # Если различие между значениями x и y двух кортежей меньше или равно порогу,
                    # помечаем exclude как True и выходим из цикла
                    exclude = True
                    break
            if not exclude:  # Если exclude равно False, добавляем кортеж в unique_list
                unique_list.append((x1_coordinate, y1_coordinate))
        matches = unique_list

        if not matches:
            self.logger.error(f"_find_many_coordinates_by_image() NO MATCHES, {image=}")
            return None

        # Добавляем правый нижний угол к каждому найденному совпадению
        matches_with_corners = []
        for match in matches:
            x_coordinate, y_coordinate = match
            width, height = small_image.shape[::-1]
            top_left = (x_coordinate, y_coordinate)
            bottom_right = (x_coordinate + width, y_coordinate + height)
            matches_with_corners.append((top_left + bottom_right))

        return matches_with_corners

    @helpers_decorators.retry
    def get_text_coordinates(
            self,
            text: str,
            image: Union[bytes, str, Image.Image, np.ndarray] = None,
            language: str = 'rus'
    ) -> Optional[tuple[int, ...]]:
        """
        Возвращает координаты области с указанным текстом на предоставленном изображении или снимке экрана.

        Аргументы:
        - text (str): Искомый текст.
        - image (bytes, str, Image.Image, np.ndarray, опционально): Изображение, на котором осуществляется поиск текста.
          Если не указано, будет использован снимок экрана. По умолчанию None.
        - language (str, опционально): Язык для распознавания текста. По умолчанию 'rus'.

        Возвращает:
        - Union[Tuple[int, int, int, int], None]: Координаты области с текстом или None, если текст не найден.
        """

        if not image:
            # Получаем снимок экрана, если изображение не предоставлено
            screenshot = self.get_screenshot_as_base64_decoded()  # Получение снимка экрана в формате base64
            image = self.to_ndarray(image=screenshot,
                                    grayscale=True)  # Преобразование снимка экрана в массив numpy и преобразование в оттенки серого
        else:
            # Если предоставлено, то преобразуем
            image = self.to_ndarray(image=image,
                                    grayscale=True)  # Преобразование изображения в массив numpy и преобразование в оттенки серого

        # Бинаризация изображения
        _, threshold = cv2.threshold(image, 0, 255,
                                     cv2.THRESH_BINARY | cv2.THRESH_OTSU)  # Применение бинаризации для получения двоичного изображения

        # Выполнение OCR с помощью PyTesseract
        data = pytesseract.image_to_data(threshold, lang=language,
                                         output_type=pytesseract.Output.DICT)  # Использование PyTesseract для распознавания текста и получения информации о распознанных словах

        formatted_data = {}
        for i in range(len(data['text'])):
            word_text = data['text'][i]  # Текст слова
            left = int(data['left'][i])  # Координата левой границы слова
            top = int(data['top'][i])  # Координата верхней границы слова
            width = int(data['width'][i])  # Ширина слова
            height = int(data['height'][i])  # Высота слова
            coordinates = [left, top, left + width, top + height]  # Координаты рамки слова

            if word_text:
                if i not in formatted_data:
                    formatted_data[i] = {}
                formatted_data[i] = {'text': word_text,
                                     'coordinates': coordinates}  # Сохранение информации о слове и его координатах

        # Разбить искомый текст на отдельные слова
        words = text.split(' ')  # Разделение искомого текста на отдельные слова

        # Инициализировать переменные для последовательности слов и соответствующих координат
        current_sequence = []  # Текущая последовательность слов
        result_coordinates = []  # Координаты текущей последовательности слов

        for word_data in formatted_data.values():
            word = word_data['text']  # Текущее слово
            coordinates = word_data['coordinates']  # Координаты слова

            if word in words:
                current_sequence.append(word)  # Добавление слова в текущую последовательность
                result_coordinates.append(coordinates)  # Добавление координат слова в результат
            else:
                if current_sequence == words:
                    # Если найдена последовательность слов, вернуть соответствующие координаты
                    top_left = tuple(map(int, result_coordinates[0][:2]))  # Верхний левый угол рамки
                    bottom_right = tuple(map(int, result_coordinates[-1][2:]))  # Нижний правый угол рамки
                    return top_left + bottom_right
                current_sequence = []  # Сброс текущей последовательности слов
                result_coordinates = []  # Сброс координат последовательности слов

        return None

    def draw_by_coordinates(self,
                            image: Union[bytes, str, Image.Image, np.ndarray] = None,
                            coordinates: Tuple[int, int, int, int] = None,
                            top_left: Tuple[int, int] = None,
                            bottom_right: Tuple[int, int] = None,
                            path: str = None) -> bool:
        """
        Рисует прямоугольник на предоставленном изображении или снимке экрана с помощью драйвера.
        Прямоугольник определяется либо координатами, либо верхней левой и нижней правой точками.
        Результирующее изображение с нарисованным прямоугольником сохраняется по указанному пути.

        Аргументы:
            image (Union[bytes, str, Image.Image], опционально): Изображение, на котором будет рисоваться.
                Может быть представлено в виде bytes, str (путь до файла) или PIL Image.
                Если не указано, используется снимок экрана. По умолчанию None.
            coordinates (Tuple[int, int, int, int], опционально): Координаты прямоугольника (x1, y1, x2, y2).
                По умолчанию None.
            top_left (Tuple[int, int], опционально): Верхняя левая точка прямоугольника (x, y). По умолчанию None.
            bottom_right (Tuple[int, int], опционально): Нижняя правая точка прямоугольника (x, y). По умолчанию None.
            path (str, опционально): Путь для сохранения результирующего изображения. По умолчанию None.

        Usage:
            image = self.driver.get_screenshot_as_base64().encode('utf-8')
            draw_by_coordinates(image=image, coordinates=(123, 123, 123, 123), path='pictures')

        Возвращает:
            bool: True, если операция выполнена успешно, False в противном случае.
        """
        try:
            if image is None:
                # Если изображение не предоставлено, получаем снимок экрана с помощью драйвера
                screenshot = self.get_screenshot_as_base64_decoded()
                image = self.to_ndarray(screenshot)
            else:
                image = self.to_ndarray(image)

            # Если верхняя левая и нижняя правая точки не предоставлены, используем координаты для определения
            # прямоугольника
            if not top_left and not bottom_right:
                top_left = (coordinates[0], coordinates[1])
                bottom_right = (coordinates[2], coordinates[3])

            # Сохраняем снимок экрана с нарисованным прямоугольником
            if path is None:
                path = "screenshot_with_text_coordinates.png"
            path = os.path.join(path)
            cv2.rectangle(image, top_left, bottom_right, (0, 255, 0), 2)
            cv2.imwrite(path, image)

            return True
        except WebDriverException as e:
            # Обрабатываем исключение WebDriverException и записываем ошибку в журнал
            self.logger.error(f'draw_by_coordinates() WebDriverException {e}')
            return False
        except cv2.error as e:
            # Обработка исключения cv2.error
            self.logger.error(f'draw_by_coordinates() cv2.error: {e}')
            return False

    @staticmethod
    def is_rgb(image: np.ndarray) -> bool:
        """
        Проверяет, является ли изображение цветным (RGB).

        Аргументы:
        - image: np.ndarray - Входное изображение в формате NumPy ndarray.

        Возвращает:
        - bool - True, если изображение является цветным (RGB), False - в противном случае.
        """
        return len(image.shape) == 3 and image.shape[2] == 3 or image.ndim == 3 or image.ndim == '3'

    @staticmethod
    def is_grayscale(image: np.ndarray) -> bool:
        """
        Проверяет, является ли изображение оттенков серого.

        Аргументы:
        - image: np.ndarray - Входное изображение в формате NumPy ndarray.

        Возвращает:
        - bool - True, если изображение является оттенков серого, False - в противном случае.
        """
        return len(image.shape) == 2 or (
                len(image.shape) == 3 and image.shape[2] == 1) or image.ndim == 2 or image.ndim == '2'

    def to_grayscale(self, image: np.ndarray) -> np.ndarray:
        """
        Преобразует изображение в оттенки серого.

        Аргументы:
        - image: np.ndarray - Входное изображение в формате ndarray.

        Возвращает:
        - np.ndarray - Преобразованное изображение в оттенках серого.
        """
        # Проверяем, является ли изображение в формате RGB
        if self.is_rgb(image):
            # Если да, то преобразуем его в оттенки серого
            gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            # Приводим значения пикселей к диапазону от 0 до 255
            gray_image = cv2.convertScaleAbs(gray_image)
            return gray_image
        # Иначе, возвращаем изображение без изменений
        return image

    def to_ndarray(self, image: Union[bytes, np.ndarray, Image.Image, str], grayscale: bool = True) -> np.ndarray:
        """
        Преобразует входные данные из различных типов в ndarray (NumPy array).

        Аргументы:
        - image: Union[bytes, np.ndarray, Image.Image, str] - Входные данные,
          представляющие изображение. Может быть типами bytes, np.ndarray, PIL Image или str.

        Возвращает:
        - np.ndarray - Преобразованный массив NumPy (ndarray) представляющий изображение.
        """
        # Если входные данные являются массивом байтов, преобразовать их в массив NumPy
        if isinstance(image, bytes):
            image = cv2.imdecode(np.frombuffer(image, np.uint8), cv2.IMREAD_COLOR)

        # Если входные данные являются строкой с путем к файлу, открыть изображение и преобразовать в массив NumPy
        if isinstance(image, str):
            # image = np.array(Image.open(image))
            image = cv2.imread(image, cv2.IMREAD_COLOR)

        # Если входные данные являются объектом PIL Image, преобразовать его в массив NumPy
        if isinstance(image, Image.Image):
            image = np.array(image)

        # Вернуть преобразованный массив NumPy
        if grayscale:
            return self.to_grayscale(image=image)
        return image



"""
Корневой класс AppiumExtended. Обеспечивает соединение с сервером и инициализацию драйвера.
"""

# coding: utf-8
import logging
import json
import time

from appium import webdriver

from appium_extended_helpers.appium_helpers import AppiumHelpers
from appium_extended_server.appium_server import AppiumServer
from appium_extended_terminal.terminal import Terminal
from appium_extended_terminal.aapt import Aapt
from appium_extended_terminal.adb import Adb


class AppiumBase:
    """
    Класс работы с Appium.
    Обеспечивает подключение к устройству
    """

    def __init__(self, logger: logging.Logger = None):
        self.server_log_level = None
        self.server_port = None
        self.server_ip = None
        self.server = None
        self.logger = logger
        self.driver = None
        self.terminal = None
        self.session_id = None
        self.helper: AppiumHelpers = None
        self.keep_alive_server = True
        self.aapt = Aapt()
        self.adb = Adb()

        aapt_logger = logging.getLogger('aapt')
        aapt_logger.setLevel(self.logger.level)
        adb_logger = logging.getLogger('adb')
        adb_logger.setLevel(self.logger.level)

    def connect(self,
                capabilities: dict,
                server_ip: str = '127.0.0.1',
                server_port: int = 4723,
                server_log_level: str = 'error',
                remote: bool = False,
                keep_alive_server: bool = True) -> None:
        """
        Подключение к устройству
        """
        self.server_ip = server_ip
        self.server_port = server_port
        self.server_log_level = server_log_level
        self.keep_alive_server = keep_alive_server
        self.server = AppiumServer(server_ip=self.server_ip,
                                   server_port=self.server_port,
                                   remote_log_level=self.server_log_level,
                                   logger=self.logger)
        self.logger.debug(
            f"connect(capabilities {capabilities}")
        if not remote:
            # запускаем локальный сервер Аппиум
            if not self.server.is_alive():
                self.server.start()
                time.sleep(10)
                self.server.wait_until_alive()

        url = f'http://{server_ip}:{str(server_port)}/wd/hub'
        self.logger.info(f"Подключение к серверу: {url}")
        self.driver = webdriver.Remote(command_executor=url,
                                       desired_capabilities=capabilities,
                                       keep_alive=True)
        self.session_id = self.driver.session_id
        # Инициализация объектов требующих драйвер
        self.terminal = Terminal(driver=self.driver, logger=self.logger)
        self.helper = AppiumHelpers(driver=self.driver, logger=self.logger)

        app_capabilities = json.dumps(capabilities)
        self.logger.info(f'Подключение установлено с  параметрами: {str(app_capabilities)}, {url}')
        self.logger.info(f'Сессия №: {self.driver.session_id}')

    def disconnect(self) -> None:
        """
        Отключение от устройства
        """
        if self.driver:
            self.logger.debug(f"Отключение от сессии №: {self.driver.session_id}")
            self.driver.quit()
            self.driver = None
        if not self.keep_alive_server:
            self.server.stop()

    def is_running(self) -> bool:
        """
        Проверяет, запущен сервер или нет
        """
        return self.driver.is_running()



# coding: utf-8
import inspect
import logging
import sys
import time
import traceback
from typing import Union, Tuple, Dict, List, Optional, cast, Any
import numpy as np
from PIL import Image

from appium.webdriver.common.appiumby import AppiumBy
from appium.webdriver.common.mobileby import MobileBy
from appium.webdriver import WebElement

from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.by import By

from appium_extended.appium_swipe import AppiumSwipe
from appium_extended.appium_wait import AppiumWait
from appium_extended.appium_tap import AppiumTap
from appium_extended.appium_is import AppiumIs

from appium_extended_web_element.web_element_extended import WebElementExtended

from appium_extended_utils import utils


class AppiumExtended(AppiumIs, AppiumTap, AppiumSwipe, AppiumWait):
    """
    Класс работы с Appium.
    Обеспечивает работу с устройством
    """

    def __init__(self, logger: logging.Logger = None, log_level: int = logging.INFO, log_path: str = ''):
        if logger is None:
            logger = logging.getLogger(__name__)
            logger.setLevel(log_level)
        if bool(log_path):
            if not log_path.endswith('.log'):
                log_path = log_path + '.log'
            file_handler = logging.FileHandler(log_path)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        super().__init__(logger=logger)

    def get_element(self,
                    locator: Union[Tuple, WebElementExtended, Dict[str, str], str] = None,
                    by: Union[MobileBy, AppiumBy, By, str] = None,
                    value: Union[str, Dict, None] = None,
                    timeout_elem: int = 10,
                    timeout_method: int = 600,
                    elements_range: Union[Tuple, List[WebElementExtended], Dict[str, str], None] = None,
                    contains: bool = True,
                    ) -> Union[WebElementExtended, None]:
        """
        Получает элемент на странице, используя указанный локатор или параметры.

        Аргументы:
        - locator (Union[Tuple, WebElementExtended, Dict[str, str], str], optional):
            Локатор или элемент для поиска. По умолчанию None.
        - by (Union[MobileBy, AppiumBy, By, str], optional): Метод поиска элемента.
            По умолчанию None.
        - value (Union[str, Dict, None], optional): Значение для поиска элемента.
            По умолчанию None.
        - timeout_elem (int, optional): Время ожидания появления элемента.
            По умолчанию 10.
        - timeout_method (int, optional): Время ожидания выполнения метода.
            По умолчанию 600.
        - elements_range (Union[Tuple, List[WebElementExtended], Dict[str, str], None], optional):
            Диапазон элементов для выбора. По умолчанию None.
        - contains (bool, optional): Флаг, указывающий на необходимость использования частичного совпадения текста.
            По умолчанию True.

        Возвращает:
        - Union[WebElementExtended, None]: Расширенный элемент WebElementExtended или None, если элемент не найден.
        """
        element = self._get_element(locator=locator,
                                    by=by,
                                    value=value,
                                    timeout_elem=timeout_elem,
                                    timeout_method=timeout_method,
                                    elements_range=elements_range,
                                    contains=contains)
        return WebElementExtended(driver=element.parent, element_id=element.id, logger=self.logger)

    def get_elements(self,
                     locator: Union[Tuple, List[WebElement], Dict[str, str], str] = None,
                     by: Union[MobileBy, AppiumBy, By, str] = None,
                     value: Union[str, Dict, None] = None,
                     timeout_elements: int = 10,
                     timeout_method: int = 600,
                     elements_range: Union[Tuple, List[WebElement], Dict[str, str], None] = None,
                     contains: bool = True,
                     ) -> Union[List[WebElementExtended], List]:
        """
        Получает элементы на странице, используя указанный локатор или параметры.

        Args:
        - locator (Union[Tuple, List[WebElement], Dict[str, str], str], optional): Локатор или элементы для поиска.
            По умолчанию None.
        - by (Union[MobileBy, AppiumBy, By, str], optional): Метод поиска элементов.
            По умолчанию None.
        - value (Union[str, Dict, None], optional): Значение для поиска элементов.
            По умолчанию None.
        - timeout_elements (int, optional): Время ожидания появления элементов.
            По умолчанию 10.
        - timeout_method (int, optional): Время ожидания выполнения метода.
            По умолчанию 600.
        - elements_range (Union[Tuple, List[WebElement], Dict[str, str], None], optional): Диапазон элементов для выбора.
            По умолчанию None.
        - contains (bool, optional): Флаг, указывающий на необходимость использования частичного совпадения текста.
            По умолчанию True.

        Returns:
        - Union[List[WebElementExtended], List]: Список расширенных элементов WebElementExtended или обычных элементов WebElement в случае ошибки.
        """
        elements = super()._get_elements(locator=locator,
                                         by=by,
                                         value=value,
                                         timeout_elements=timeout_elements,
                                         timeout_method=timeout_method,
                                         elements_range=elements_range,
                                         contains=contains)
        elements_ext = []
        for element in elements:
            elements_ext.append(WebElementExtended(driver=element.parent, element_id=element.id, logger=self.logger))
        return elements_ext

    def get_image_coordinates(self,
                              image: Union[bytes, np.ndarray, Image.Image, str],
                              full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                              threshold: float = 0.7,
                              ) -> Union[Tuple, None]:
        return self._get_image_coordinates(full_image=full_image,
                                           image=image,
                                           threshold=threshold)

    def get_inner_image_coordinates(self,
                                    outer_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                    inner_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                    threshold: Optional[float] = 0.9
                                    ) -> Union[Tuple[int, int, int, int], None]:
        """
        Возвращает координаты x y (относительно экрана) по изображению внутри другого изображения
        """
        return self._get_inner_image_coordinates(outer_image_path=outer_image_path,
                                                 inner_image_path=inner_image_path,
                                                 threshold=threshold)

    def get_many_coordinates_of_image(self,
                                      image: Union[bytes, np.ndarray, Image.Image, str],
                                      full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                                      cv_threshold: Optional[float] = 0.7,
                                      coord_threshold: Optional[int] = 5,
                                      ) -> Union[List[Tuple], None]:
        return self.helper.get_many_coordinates_of_image(full_image=full_image,
                                                        image=image,
                                                        cv_threshold=cv_threshold,
                                                        coord_threshold=coord_threshold)

    def get_text_coordinates(self,
                             text: str,
                             language: Optional[str] = 'rus',
                             image: Union[bytes, str, Image.Image, np.ndarray] = None,
                             ocr: Optional[bool] = True,
                             ) -> Union[Tuple[int, int, int, int], None]:  # TODO реализовать None
        """
        # TODO fill
        """
        if ocr:
            return self._get_text_coordinates(text=text, language=language, image=image)
        return self.get_element(locator={'text': text, 'displayed': 'true', 'enabled': 'true'}).get_coordinates()

    # DOM

    def get_element_contains(self,
                             ) -> Any:
        """
        Возвращает элемент содержащий определенный элемент
        """
        raise NotImplementedError("This method is not implemented yet.")  # TODO implement

    def get_elements_contains(self,
                              ) -> Any:
        """
        Возвращает элементы содержащие определенный(е) элемент(ы)
        """
        raise NotImplementedError("This method is not implemented yet.")  # TODO implement

    def get_screenshot_as_base64_decoded(self) -> bytes:
        return self._get_screenshot_as_base64_decoded()

    def find_and_get_element(self,
                             locator: Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str],
                             timeout: int = 10,
                             ) -> Union[WebElementExtended, None]:
        """
        Ищет элемент на странице, если нет то скроллит все что скроллится и ищет там
        """
        if self.is_element_within_screen(locator=locator, timeout=1):
            return self.get_element(locator=locator, timeout_elem=timeout)
        recyclers = self.get_elements(locator={'scrollable': 'true', 'enabled': 'true', 'displayed': 'true'})
        for recycler in recyclers:
            try:
                if recycler.scroll_until_find(locator=locator):
                    return self.get_element(locator=locator, timeout_elem=timeout)
            except StaleElementReferenceException as e:
                current_function_name = inspect.currentframe().f_globals['__name__']
                self.logger.error(f"{current_function_name} ERROR: {e}")
                self.logger.error(f"arg {recycler=}")
                self.logger.error(f"arg {locator=}")
                traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
                self.logger.error(traceback_info)
        return None

    def is_element_within_screen(self,
                                 locator: Union[Tuple, WebElement, 'WebElementExtended', Dict[str, str], str],
                                 timeout: int = 10,
                                 ) -> bool:
        """
        Метод проверяет, находится ли заданный элемент на видимом экране.

        Аргументы:
        - locator (Union[tuple, WebElement]): Локатор или элемент, который нужно проверить.
        - timeout (int): Время ожидания элемента. Значение по умолчанию: 10.

        Возвращает:
        - bool: True, если элемент находится на экране, False, если нет.
        """
        return self._is_element_within_screen(locator=locator, timeout=timeout)

    def is_text_on_screen(self,
                          text: str,
                          language: str = 'rus',
                          ocr: bool = True,
                          contains: bool = True
                          ) -> bool:
        """
        Проверяет, присутствует ли заданный текст на экране.
        Если ocr=True:
            Распознавание текста производит с помощью библиотеки pytesseract.
        Если ocr=False:
            Производится поиск элемента по xpath.

        Аргументы:
        - text (str): Текст, который нужно найти на экране.
        - ocr (bool): Производить поиск по изображению или DOM.
        - language (str): Язык распознавания текста. Значение по умолчанию: 'rus'.
        - contains (bool): Только для ocr=False. Допускает фрагмент текста

        Возвращает:
        - bool: True, если заданный текст найден на экране. False в противном случае.
        """
        if ocr:
            return self.helper.is_text_on_ocr_screen(text=text, language=language)
        return self._is_element_within_screen(locator={'text': text}, contains=contains)

    def is_image_on_the_screen(self,
                               image: Union[bytes, np.ndarray, Image.Image, str],
                               threshold: float = 0.9,
                               ) -> bool:
        return self.helper.is_image_on_the_screen(image=image,
                                                  threshold=threshold)

    def tap(self,
            locator: Union[Tuple[str, str], WebElementExtended, WebElement, Dict[str, str], str] = None,
            x: int = None,
            y: int = None,
            image: Union[bytes, np.ndarray, Image.Image, str] = None,
            duration: Optional[int] = None,
            timeout: int = 5,
            ) -> Union['AppiumExtended', None]:
        """
        Тап по координатам / элементу / изображению
        """
        if locator is not None:
            # Извлечение координат
            x, y = self._extract_point_coordinates_by_typing(locator)
        if image is not None:
            start_time = time.time()
            while not self.is_image_on_the_screen(image=image) and time.time() - start_time < timeout:
                time.sleep(1)
            # Извлечение координат
            x, y = self._extract_point_coordinates_by_typing(image)

        assert self._tap(x=x, y=y,
                         duration=duration)
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    # SWIPE
    def swipe(self,
              start_position: Union[
                  Tuple[int, int], str, bytes, np.ndarray, Image.Image, WebElement, WebElementExtended, Tuple[str, str],
                  Dict[str, str]],
              end_position: Optional[Union[
                  Tuple[int, int], str, bytes, np.ndarray, Image.Image, WebElement, WebElementExtended, Tuple[str, str],
                  Dict[str, str]]] = None,
              direction: Optional[int] = None,
              distance: Optional[int] = None,
              duration: Optional[int] = 0,
              ) -> 'AppiumExtended':
        """
        Выполняет свайп (перетаскивание) элемента или изображения на экране.

        Параметры:
        - start_position: Позиция начала свайпа. Может быть задана в различных форматах:
            - Если `start_position` является кортежем и оба его элемента являются строками, то он представляет собой
              локатор элемента. В этом случае будет выполнен поиск элемента и используется его позиция.
            - Если `start_position` является словарем, то считается, что это локатор элемента, основанный на атрибутах.
              Например, {'text': 'some text'} или {'class': 'SomeClass', 'visible': 'true'}. В этом случае будет
              выполнен поиск элемента по указанным атрибутам, и используется его позиция.
            - Если `start_position` является экземпляром класса WebElement или WebElementExtended, то используется его
              позиция.
            - Если `start_position` является строкой, массивом байтов (bytes), массивом NumPy (np.ndarray) или объектом
              класса Image.Image, то считается, что это изображение. В этом случае будет вычислен центр изображения и
              используется его позиция.
            - Если `start_position` является кортежем, и оба его элемента являются целыми числами, то считается, что это
              координаты в формате (x_coordinate, y_coordinate).

        - end_position: Позиция конца свайпа. Принимает те же форматы, что и `start_position`. По умолчанию None.
        - direction (опционально): Направление свайпа. Принимает значения от 0 до 360 градусов.
          Если указано направление, то будет вычислена конечная точка свайпа на основе текущего размера окна и
          указанного расстояния. По умолчанию None.
        - distance (опционально): Расстояние свайпа. Принимается в пикселях. Используется только в сочетании с параметром
          `direction`. По умолчанию None.
        - duration (опционально): Продолжительность свайпа в миллисекундах. По умолчанию 0.

        Возвращает:
        - self: Экземпляр класса appium_extended.

        Примечания:
        - В качестве конечной позиции свайпа должен быть указан end_position или пара direction, distance.
        - str принимается как путь к изображению на экране и вычисляется его центр, а не как локатор элемента

        """
        # Извлечение координат начальной точки свайпа
        start_x, start_y = self._extract_point_coordinates_by_typing(start_position)

        if end_position is not None:
            # Извлечение координат конечной точки свайпа
            end_x, end_y = self._extract_point_coordinates_by_typing(end_position)
        else:
            # Извлечение координат конечной точки свайпа на основе направления и расстояния
            end_x, end_y = self._extract_point_coordinates_by_direction(direction, distance, start_x, start_y,
                                                                        screen_resolution=self.terminal.get_screen_resolution())

        # Выполнение свайпа
        assert self._swipe(start_x=start_x, start_y=start_y,
                           end_x=end_x, end_y=end_y,
                           duration=duration)

        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    def swipe_right_to_left(self) -> 'AppiumExtended':
        window_size = self.terminal.get_screen_resolution()
        width = window_size[0]
        height = window_size[1]
        left = int(width * 0.1)
        right = int(width * 0.9)
        self.swipe(start_position=(right, height // 2),
                   end_position=(left, height // 2))
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    def swipe_left_to_right(self) -> 'AppiumExtended':
        window_size = self.terminal.get_screen_resolution()
        width = window_size[0]
        height = window_size[1]
        left = int(width * 0.1)
        right = int(width * 0.9)
        self.swipe(start_position=(left, height // 2),
                   end_position=(right, height // 2))
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    def swipe_top_to_bottom(self) -> 'AppiumExtended':
        window_size = self.terminal.get_screen_resolution()
        height = window_size[1]
        top = int(height * 0.1)
        bottom = int(height * 0.9)
        self.swipe(start_position=(top, height // 2),
                   end_position=(bottom, height // 2))
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    def swipe_bottom_to_top(self) -> 'AppiumExtended':
        window_size = self.terminal.get_screen_resolution()
        height = window_size[1]
        top = int(height * 0.1)
        bottom = int(height * 0.9)
        self.swipe(start_position=(bottom, height // 2),
                   end_position=(top, height // 2))
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    # WAIT

    def wait_for(self,
                 locator: Union[Tuple[str, str], WebElement, 'WebElementExtended', Dict[str, str], str,
                 List[Tuple[str, str]], List[WebElement], List['WebElementExtended'], List[Dict[str, str]], List[
                     str]] = None,
                 image: Union[bytes, np.ndarray, Image.Image, str,
                 List[bytes], List[np.ndarray], List[Image.Image], List[str]] = None,
                 timeout: int = 10,
                 contains: bool = True,
                 full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                 ) -> 'AppiumExtended':
        assert self._wait_for(locator=locator,
                              image=image,
                              timeout=timeout,
                              contains=contains)
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    def wait_for_not(self,
                     locator: Union[Tuple[str, str], WebElement, 'WebElementExtended', Dict[str, str], str,
                     List[Tuple[str, str]], List[WebElement], List['WebElementExtended'], List[Dict[str, str]], List[
                         str]] = None,
                     image: Union[bytes, np.ndarray, Image.Image, str,
                     List[bytes], List[np.ndarray], List[Image.Image], List[str]] = None,
                     timeout: int = 10,
                     contains: bool = True,
                     ) -> 'AppiumExtended':
        assert self._wait_for_not(locator=locator, image=image, timeout=timeout, contains=contains)
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    def wait_return_true(self, method, timeout: int = 10) -> 'AppiumExtended':
        assert self._wait_return_true(method=method, timeout=timeout)
        # Возвращаем экземпляр класса appium_extended
        return cast('AppiumExtended', self)

    # KEYBOARD

    def input_by_virtual_keyboard(self) -> 'appium_extended':  # TODO реализовать возврат cast('appium_extended', self)
        """
        Вводит с помощью виртуально клавиатуры
        """
        raise NotImplementedError("This method is not implemented yet.")

    # OTHER

    def draw_by_coordinates(self,
                            image: Union[bytes, str, Image.Image, np.ndarray] = None,
                            coordinates: Tuple[int, int, int, int] = None,
                            top_left: Tuple[int, int] = None,
                            bottom_right: Tuple[int, int] = None,
                            path: str = None,
                            ) -> 'AppiumExtended':
        assert self.helper.draw_by_coordinates(image=image,
                                              coordinates=coordinates,
                                              top_left=top_left,
                                              bottom_right=bottom_right,
                                              path=path)
        return cast('AppiumExtended', self)

    # PRIVATE

    def _extract_point_coordinates_by_typing(self,
                                             position:
                                             Union[Tuple[int, int], str, bytes, np.ndarray, Image.Image,
                                             Tuple[str, str], Dict, WebElement, WebElementExtended]
                                             ) -> Tuple[int, int]:
        """
        Извлекает координаты точки на основе типа переданной позиции.

        Параметры:
            position (
            Union[Tuple[int, int], str, bytes, np.ndarray, Image.Image, Dict, WebElement, WebElementExtended]
            ):
                Позиция, для которой нужно извлечь координаты.
                Либо локатор элемента, либо изображение, либо кортеж из координат.

        Возвращает:
            Tuple[int, int]: Кортеж координат точки, в формате (x, y).
        """
        x, y = 0, 0
        # Вычисление позиции начала свайпа
        if (isinstance(position, Tuple) and
                isinstance(position[0], int) and
                isinstance(position[1], int)):
            # Если position является кортежем с двумя целыми числами, то считаем, что это координаты
            x, y = position
        elif (isinstance(position, Tuple) and
              isinstance(position[0], str) and
              isinstance(position[1], str)) or \
                isinstance(position, WebElement) or \
                isinstance(position, WebElementExtended) or \
                isinstance(position, Dict):
            # Если position является кортежем с двумя строковыми элементами или экземпляром WebElement,
            # WebElementExtended или словарем, то получаем координаты центра элемента
            x, y = utils.calculate_center_of_coordinates(
                self.get_element(locator=position).get_coordinates())
        elif isinstance(position, (bytes, np.ndarray, Image.Image, str)):
            # Если position является строкой, байтами, массивом NumPy или объектом Image.Image,
            # то получаем координаты центра изображения
            x, y = utils.calculate_center_of_coordinates(
                self.get_image_coordinates(image=position))
        return x, y

    @staticmethod
    def _extract_point_coordinates_by_direction(direction: int, distance: int,
                                                start_x: int, start_y: int,
                                                screen_resolution: tuple
                                                ) -> Tuple[int, int]:
        """
        Извлекает координаты точки на заданном расстоянии и в заданном направлении относительно начальных координат.

        Параметры:
            direction (str): Направление движения в пределах 360 градусов.
            distance (int): Расстояние, на которое нужно переместиться относительно начальных координат в пикселях.
            start_x (int): Начальная координата X.
            start_y (int): Начальная координата Y.

        Возвращает:
            Tuple[int, int]: Координаты конечной точки в формате (x, y).
        """
        width = screen_resolution[0]
        height = screen_resolution[1]
        end_x, end_y = utils.find_coordinates_by_vector(width=width, height=height,
                                                        direction=direction, distance=distance,
                                                        start_x=start_x, start_y=start_y)
        return end_x, end_y



# coding: utf-8
"""
Этот модуль содержит полезные декораторы для работы с Appium.
"""
import base64
import sys
import io
import time
import functools
import traceback
from functools import wraps
from datetime import datetime

import allure
import numpy as np
from PIL import Image


# TODO make unit test for module

def retry(func):
    """
    Повторяет выполнение метода если он возвращает False или None.
    3 tries hardcode in method
    """
    max_retries = 3

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = None
        for _ in range(max_retries):
            result = func(*args, **kwargs)
            if result is not None and result is not False:
                return result
            time.sleep(1)
        return result

    # Возвращаем обертку функции
    return wrapper


def wait_until_window_change(poll_frequency: float = 0.1):
    """
    Декоратор, который ожидает пока окно не перестанет меняться.
    В обернутом методе должен быть аргумент
        decorator_args: dict
            timeout_window: общее время ожидания
            window_not_changing_period: период времени в течении которого окно не должно изменятся

    Аргументы:
        poll_frequency (float): Частота опроса содержимого окна на наличие изменений в секундах.
                               По умолчанию 0.1 секунды.

    Возвращает:
        function: Декорированная функция.
    """

    def func_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Оберточная функция, которая инкапсулирует декорированную функцию
            с логикой обнаружения изменений окна.

            Аргументы:
                self: Экземпляр класса, к которому принадлежит
                      декорированный метод.
                *args: Произвольное число аргументов, переданных в
                       декорированный метод.
                **kwargs: Произвольное число именованных аргументов, переданных в
                          декорированный метод.

            Возвращает:
                bool: True, если содержимое окна изменяется в течении
                      заданного периода времени, иначе False.
            """

            # Инициализация
            func_result = False
            decorator_args = kwargs.get('decorator_args', {})
            timeout_window = decorator_args.get('timeout_window', 30)
            window_not_changing_period = decorator_args.get('window_not_changing_period', 10)
            # Запись начального времени
            start_time = time.time()
            # Вызов декорированной функции и сохранение результата
            func_result = func(self, *args,
                               **kwargs)

            # Обнаружение изменений экрана с экспоненциальной задержкой
            poll_interval = poll_frequency
            # Продолжаем до достижения тайм-аута
            while time.time() - start_time < timeout_window:
                # Запускаем новый период, в течение которого окно не изменяется
                window_not_changing_period_start_time = time.time()
                # Флаг для отслеживания того, изменилось ли окно за период
                window_not_changed = True

                while (time.time() - window_not_changing_period_start_time
                       < window_not_changing_period):
                    # Делаем снимок экрана и сохраняем его в памяти
                    image_bytes = self.driver.get_screenshot_as_png()
                    # Преобразуем в оттенки серого
                    image = Image.open(io.BytesIO(image_bytes)).convert('L')
                    # Обрезаем снимок до определенной области (лево, верх, право, низ)
                    box = (50, 50, 400, 400)
                    image = image.crop(box)
                    # Ждем указанный интервал между опросами
                    time.sleep(poll_interval)
                    new_image_bytes = self.driver.get_screenshot_as_png()
                    # Преобразуем в оттенки серого
                    new_image = Image.open(io.BytesIO(new_image_bytes)).convert('L')
                    new_image = new_image.crop(box)

                    # Проверяем, отличается ли сумма значений пикселей на двух изображениях
                    if np.sum(image) != np.sum(new_image):
                        # Содержимое окна изменилось
                        window_not_changed = False
                        break

                if window_not_changed:
                    self.logger.debug("Содержимое окна не изменялось в течение периода")
                    return True
                # Удваиваем время ожидания для каждого опроса
                poll_interval *= 2

            if not func_result:
                self.logger.info(f"{func.__name__}() > {func_result}. Изменение экрана: False")
                return False

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def wait_for_window_change(poll_frequency: float = 0.5):
    """
    Декоратор, который ожидает изменения окна после выполнения метода.
    Если окно не изменилось - выполняет еще попытку.
    В обернутом методе должен быть аргумент
        decorator_args: dict
            timeout_window: int, время ожидания на попытку
            tries: int, количество попыток выполнения метода

    Аргументы:
        poll_frequency (float): Частота проверки окна на изменения.

    Возвращает:
        Декоратор функции.
    """

    def func_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Оберточная функция, которая выполняет обнаружение изменения окна и
            выполнение декорированной функции.

            Аргументы:
                self: Экземпляр класса, содержащего декорированную функцию.
                *args: Переменное число аргументов.
                **kwargs: Произвольные именованные аргументы.

            Возвращает:
                Результат декорированной функции или False, если изменение окна не было обнаружено.
            """

            # Инициализация
            func_result = False
            decorator_args = kwargs.get('decorator_args', {})
            timeout_window = decorator_args.get('timeout_window', 10)
            tries = decorator_args.get('tries', 3)

            # Сделать снимок экрана и сохранить его в памяти в виде байтов
            image_bytes = self.driver.get_screenshot_as_png()
            # Открыть изображение из байтов и преобразовать его в оттенки серого
            image = Image.open(io.BytesIO(image_bytes)).convert(
                'L')

            # Обрезать снимок экрана до определенной области (лево, верх, право, низ)
            box = (50, 50, 400, 400)
            # Обрезать изображение на основе заданных координат области
            image = image.crop(box)

            # Попытаться обнаружить изменение экрана в течении tries попыток
            for _ in range(tries):
                # Записать текущее время начала попытки обнаружения
                start_time = time.time()
                # Выполнить декорированную функцию и сохранить результат
                func_result = func(self, *args,
                                   **kwargs)

                # Обнаружить изменение экрана с экспоненциальной задержкой
                poll_interval = poll_frequency
                # Проверить, находится ли прошедшее время в пределах заданного окна времени ожидания
                while time.time() - start_time < timeout_window:
                    # Приостановить выполнение на заданный интервал проверки
                    time.sleep(poll_interval)
                    # Сделать новый снимок экрана окна и получить данные изображения в виде байтов
                    new_image_bytes = self.driver.get_screenshot_as_png()
                    # Открыть новое изображение из байтов и преобразовать его в оттенки серого
                    new_image = Image.open(io.BytesIO(new_image_bytes)).convert(
                        'L')
                    # Обрезать новое изображение на основе заданных координат области
                    new_image = new_image.crop(
                        box)
                    # Сравнить суммы значений пикселей между исходным и новым изображениями
                    if not np.sum(image) == np.sum(
                            new_image):
                        # Записать сообщение о том, что произошло изменение экрана
                        self.logger.debug(
                            "Изменение экрана обнаружено")
                        # Вернуть True для обозначения обнаружения изменения экрана
                        return True
                        # Удвоить интервал проверки для следующей итерации
                        # (экспоненциальная задержка)
                    poll_interval *= 2

            if not func_result:
                # Записать сообщение о том, что изменение экрана не было обнаружено
                self.logger.info(
                    f"{func.__name__}() > {func_result}. Изменение экрана: False")
                # Вернуть False для обозначения отсутствия обнаружения изменения экрана
                return False

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def time_it(func):
    """
    Замеряет время выполнения метода.
    Печатает результат замера.
    """

    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        print(f"Execution time of {func.__name__}: {execution_time:.2f} seconds")
        return result

    # Возвращаем обертку функции
    return wrapper


def step_info(my_str):
    """
    Декоратор, который перед вызовом метода вызывает logger.info и @allure.step,
    передавая в них строковую переменную, принятую в параметрах.

    Аргументы:
        my_str (str): Строковая переменная для использования в logger.info и @allure.step.

    Возвращает:
        function: Декоратор функций.

    Пример использования:
        @my_step_info("Мой шаг")
        def my_function():
            ...
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        @allure.step(my_str)
        def wrapper(self, *args, **kwargs):
            result = None
            # Логируем информацию перед вызовом метода
            self.logger.info(my_str)
            # Получаем скриншот до вызова метода
            screenshot = self.driver.get_screenshot_as_png()
            # Генерируем временную метку для имени скриншота
            timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            # Устанавливаем имя скриншота до вызова метода
            screenshot_name_begin = f"screenshot_begin_{timestamp}.png"
            # Имя файла видеозаписи с временной меткой
            video_filename = f'screenrecord_{timestamp}.mp4'
            self.driver.start_recording_screen()
            try:
                # Выполняем исходную функцию
                result = func(self, *args, **kwargs)
            except Exception as error:
                # Если произошло исключение, прикрепляем скриншот до вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_begin,
                              attachment_type=allure.attachment_type.PNG)
                # Получаем скриншот после вызова метода
                screenshot = self.driver.get_screenshot_as_png()
                # Генерируем временную метку для имени скриншота
                timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
                # Устанавливаем имя скриншота до вызова метода
                screenshot_name_end = f"screenshot_end_{timestamp}.png"
                # Если произошло исключение, прикрепляем скриншот после вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_end,
                              attachment_type=allure.attachment_type.PNG)

                # Если произошло исключение, прикрепляем видеозапись выполнения метода к отчету
                allure.attach(base64.b64decode(self.driver.stop_recording_screen()),
                              name=video_filename,
                              attachment_type=allure.attachment_type.MP4)

                # Прикрепляем информацию об ошибке AssertionError к отчету
                allure.attach(str(error),
                              name="AssertionError",
                              attachment_type=allure.attachment_type.TEXT)

                # Выводим информацию в лог
                self.logger.error(f"{my_str} [не выполнено]")
                self.logger.error("AssertionError:")
                traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
                self.logger.error(traceback_info)

                # Прокидываем исключение дальше
                raise AssertionError(str(error)).with_traceback(sys.exc_info()[2])

            # Логируем информацию после успешного выполнения метода
            self.logger.info(f"{my_str} [выполнено успешно]")
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def screenshots():
    """
    В случае возникновения AssertionError в обернутом методе -
    прикрепляет к allure report скриншот до выполнения
    метода и после возникновения исключения, а также информацию об ошибке.
    В ином случае скриншот не прикрепляется.
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            # Получаем скриншот до вызова метода
            screenshot = self.driver.get_screenshot_as_png()
            # Генерируем временную метку для имени скриншота
            timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            # Устанавливаем имя скриншота до вызова метода
            screenshot_name_begin = f"screenshot_begin_{timestamp}.png"
            try:
                # Выполняем исходную функцию
                result = func(self, *args, **kwargs)
            except AssertionError as error:
                # Если произошло исключение, прикрепляем скриншот до вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_begin,
                              attachment_type=allure.attachment_type.PNG)
                # Прикрепляем информацию об ошибке AssertionError к отчету
                allure.attach(str(error),
                              name="AssertionError",
                              attachment_type=allure.attachment_type.TEXT)
                # Рейзим исключение AssertionError с сохраненным traceback
                raise AssertionError(str(error)).with_traceback(sys.exc_info()[2])
            finally:
                # Получаем скриншот после вызова метода
                screenshot = self.driver.get_screenshot_as_png()
                # Обновляем временную метку для имени скриншота
                timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
                # Устанавливаем имя скриншота после вызова метода
                screenshot_name_end = f"screenshot_end_{timestamp}.png"
                # Прикрепляем скриншот после вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_end,
                              attachment_type=allure.attachment_type.PNG)
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def log_debug():
    """
    Логирует начало и завершение обернутого метода
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            # Получаем имя метода
            method_name = func.__name__
            # Логируем начало выполнения метода и переданные аргументы
            self.logger.debug(f"{method_name}() < {', '.join(map(str, args))}, "
                              f"{', '.join(f'{k}={v}' for k, v in kwargs.items())}")
            # Выполняем исходную функцию
            result = func(self, *args, **kwargs)
            # Если результат существует, логируем его
            if result:
                self.logger.debug(f"{method_name}() > {str(result)}")
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def print_me():
    """
    Печатает начало и завершение обернутого метода
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        def wrapper(*args, **kwargs):
            # Получаем имя метода
            method_name = func.__name__
            # Печатаем начало выполнения метода и переданные аргументы
            print(f"{method_name}() < {', '.join(map(str, args))}, "
                  f"{', '.join(f'{k}={v}' for k, v in kwargs.items())}")
            # Выполняем исходную функцию
            result = func(*args, **kwargs)
            # Если результат существует, логируем его
            if result:
                print(f"{method_name}() > {str(result)}")
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator



import subprocess

from setuptools import setup, find_packages

# Общее описание пакета
long_description = '''
An extension library for adding ease of use to Appium-Python-Client.

appium_extended is a collection of utilities and convenience functions designed to enhance the usage of Appium-Python-Client for mobile app automation testing. It provides additional functionalities and abstractions to simplify the testing process.
'''

VERSION = '0.1.33'


def install_requirements():
    # Первые зависимости, которые должны быть установлены перед Pillow
    pre_pillow_dependencies = [
        "zlib-compress==0.0.1",
        "zlib-decompress==0.0.2",
        "pylibjpeg==1.4.0"
    ]

    for package in pre_pillow_dependencies:
        subprocess.check_call(["pip", "install", package])

    # Очистка кеша перед установкой Pillow
    subprocess.check_call(["pip", "cache", "purge"])


install_requirements()

setup(
    name='AppiumExtended',
    version=VERSION,
    description='An extension library for adding ease of use Appium-Python-Client',
    author='molokov-klim',
    packages=find_packages(),
    install_requires=[
        'Pillow>=9.5.0',
        'Appium-Python-Client>=2.11.1',
        'allure-pytest>=2.13.2',
        'requests>=2.31.0',
        'pyserial>=3.5',
        'opencv-python>=4.8.0.74',
        'pytesseract>=0.3.10',
        'numpy>=1.25.1',
        'selenium>=4.10.0',
    ],
    # Добавляем описание в setup()
    long_description=long_description,
    long_description_content_type='text/plain',  # Указываем тип контента (обычный текст)
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    url='https://github.com/molokov-klim/appium_extended',
)
